import importlib
import os
import sys
from pathlib import Path
from typing import List, MutableMapping, Optional, Tuple, Union

import yaml
from infrahub_sdk.schema import GenericSchema, NodeSchema

from diffsync.store.local import LocalStore
from diffsync.store.redis import RedisStore
from infrahub_sync import SyncAdapter, SyncConfig, SyncInstance
from infrahub_sync.generator import render_template
from potenda import Potenda


def render_adapter(
    sync_instance: SyncInstance, schema: MutableMapping[str, Union[NodeSchema, GenericSchema]]
) -> List[Tuple[str, str]]:
    files_to_render = (
        ("diffsync_models.j2", "sync_models.py"),
        ("diffsync_adapter.j2", "sync_adapter.py"),
    )
    rendered_files = []
    for adapter in [sync_instance.source, sync_instance.destination]:
        output_dir_absolute = str(os.path.join(sync_instance.directory, adapter.name))

        output_dir_path = Path(output_dir_absolute)
        if not output_dir_path.is_dir():
            output_dir_path.mkdir(exist_ok=True)

        init_file_path = output_dir_path / "__init__.py"
        if not init_file_path.exists():
            init_file_path.touch()

        for item in files_to_render:
            render_template(
                template_file=item[0],
                output_dir=output_dir_absolute,
                output_file=item[1],
                context={"schema": schema, "adapter": adapter, "config": sync_instance},
            )
            output_file_path = Path(output_dir_absolute, item[1])
            rendered_files.append((item[0], str(output_file_path)))

    return rendered_files


def import_adapter(sync_instance: SyncInstance, adapter: SyncAdapter):
    directory = Path(sync_instance.directory)
    sys.path.insert(0, str(directory))
    adapter_file_path = directory / f"{adapter.name}" / "sync_adapter.py"

    try:
        adapter_name = f"{adapter.name.title()}Sync"
        spec = importlib.util.spec_from_file_location(f"{adapter.name}.adapter", str(adapter_file_path))
        adapter_module = importlib.util.module_from_spec(spec)
        sys.modules[f"{adapter.name}.adapter"] = adapter_module
        spec.loader.exec_module(adapter_module)

        adapter_class = getattr(adapter_module, adapter_name, None)
        if adapter_class is None:
            raise AttributeError(f"{adapter_name} not found in adapter.py")
    except (FileNotFoundError, AttributeError) as exc:
        raise ImportError(f"{adapter_name}: {str(exc)}") from exc
    return adapter_class


def get_all_sync(directory: Optional[str] = None) -> List[SyncInstance]:
    results = []
    search_directory = Path(directory) if directory else Path(__file__).parent
    config_files = search_directory.glob("**/config.yml")

    for config_file in config_files:
        with config_file.open("r") as file:
            directory_name = str(config_file.parent)
            config_data = yaml.safe_load(file)
            SyncConfig(**config_data)
            results.append(SyncInstance(**config_data, directory=directory_name))

    return results


def get_instance(
    name: Optional[str] = None, config_file: Optional[str] = "config.yml", directory: Optional[str] = None
) -> Optional[SyncInstance]:
    if name:
        all_sync_instances = get_all_sync(directory=directory)
        for item in all_sync_instances:
            if item.name == name:
                return item

    config_file_path = None
    if Path(config_file).is_absolute() or directory is None:
        config_file_path = Path(config_file)
    elif directory:
        config_file_path = Path(directory) / Path(config_file)

    if config_file_path:
        directory_path = config_file_path.parent
        if config_file_path.is_file():
            with config_file_path.open("r", encoding="UTF-8") as file:
                config_data = yaml.safe_load(file)
                return SyncInstance(**config_data, directory=str(directory_path))

    return None


def get_potenda_from_instance(
    sync_instance: SyncInstance, branch: Optional[str] = None, show_progress: Optional[bool] = True
) -> Potenda:
    source = import_adapter(sync_instance=sync_instance, adapter=sync_instance.source)
    destination = import_adapter(sync_instance=sync_instance, adapter=sync_instance.destination)

    source_store = LocalStore()
    destination_store = LocalStore()

    if sync_instance.store:
        if sync_instance.store.type == "redis":
            if sync_instance.store.settings and isinstance(sync_instance.store.settings, dict):
                redis_settings = sync_instance.store.settings
                source_store = RedisStore(**redis_settings, name=sync_instance.source.name)
                destination_store = RedisStore(**redis_settings, name=sync_instance.destination.name)
            else:
                source_store = RedisStore(name=sync_instance.source.name)
                destination_store = RedisStore(name=sync_instance.destination.name)

    if sync_instance.source.name == "infrahub" and branch:
        src = source(
            config=sync_instance,
            target="source",
            adapter=sync_instance.source,
            branch=branch,
            internal_storage_engine=source_store,
        )
    else:
        src = source(
            config=sync_instance,
            target="source",
            adapter=sync_instance.source,
            internal_storage_engine=source_store,
        )
    if sync_instance.destination.name == "infrahub" and branch:
        dst = destination(
            config=sync_instance,
            target="destination",
            adapter=sync_instance.destination,
            branch=branch,
            internal_storage_engine=destination_store,
        )
    else:
        dst = destination(
            config=sync_instance,
            target="destination",
            adapter=sync_instance.destination,
            internal_storage_engine=destination_store,
        )

    ptd = Potenda(
        destination=dst,
        source=src,
        config=sync_instance,
        top_level=sync_instance.order,
        show_progress=show_progress,
    )

    return ptd



from typing import Any, Dict, List, Optional

try:
    from pydantic import v1 as pydantic  # type: ignore[attr-defined]
except ImportError:
    import pydantic  # type: ignore[no-redef]


class SchemaMappingField(pydantic.BaseModel):
    name: str
    mapping: Optional[str] = pydantic.Field(default=None)
    static: Optional[Any] = pydantic.Field(default=None)
    reference: Optional[str] = pydantic.Field(default=None)


class SchemaMappingModel(pydantic.BaseModel):
    name: str
    mapping: str
    identifiers: Optional[List[str]] = pydantic.Field(default=None)
    fields: List[SchemaMappingField] = []


class SyncAdapter(pydantic.BaseModel):
    name: str
    settings: Optional[Dict[str, Any]] = {}


class SyncStore(pydantic.BaseModel):
    type: str
    settings: Optional[Dict[str, Any]] = {}


class SyncConfig(pydantic.BaseModel):
    name: str
    store: Optional[SyncStore] = []
    source: SyncAdapter
    destination: SyncAdapter
    order: List[str] = pydantic.Field(default_factory=list)
    schema_mapping: List[SchemaMappingModel] = []


class SyncInstance(SyncConfig):
    directory: str


class DiffSyncMixin:
    def load(self):
        """Load all the models, one by one based on the order defined in top_level."""
        for item in self.top_level:
            if hasattr(self, f"load_{item}"):
                print(f"Loading {item}")
                method = getattr(self, f"load_{item}")
                method()
            else:
                print(f"Loading {item}")
                self.model_loader(model_name=item, model=getattr(self, item))

    def model_loader(self, model_name: str, model):
        raise NotImplementedError


class DiffSyncModelMixin:
    @classmethod
    def is_list(cls, name):
        field = cls.__fields__.get(name)
        if not field:
            raise ValueError(f"Unable to find the field {name} under {cls}")

        if isinstance(field.default, list):
            return True

        return False



from __future__ import annotations

# pylint: disable=R0801
import os
from typing import TYPE_CHECKING, Any, Dict

import pynetbox

from diffsync import DiffSyncModel
from infrahub_sync import (
    DiffSyncMixin,
    DiffSyncModelMixin,
    SchemaMappingModel,
    SyncAdapter,
    SyncConfig,
)

try:
    from diffsync import Adapter as DiffSync  # type: ignore[attr-defined]
except ImportError:
    from diffsync import DiffSync  # type: ignore[no-redef]

if TYPE_CHECKING:
    from pynetbox.core.response import Record as NetboxRecord


def get_value(obj, name: str):
    """Query a value in dot notation recursively on a NetboxRecord"""
    if "." not in name:
        return getattr(obj, name)

    first_name, remaining_part = name.split(".", maxsplit=1)
    sub_obj = getattr(obj, first_name)
    if not sub_obj:
        return None
    return get_value(obj=sub_obj, name=remaining_part)


class NetboxAdapter(DiffSyncMixin, DiffSync):
    type = "Netbox"

    def __init__(self, *args, target: str, adapter: SyncAdapter, config: SyncConfig, **kwargs):
        super().__init__(*args, **kwargs)

        self.target = target
        self.client = self._create_netbox_client(adapter)
        self.config = config

    def _create_netbox_client(self, adapter: SyncAdapter):
        settings = adapter.settings or {}
        url = os.environ.get("NETBOX_ADDRESS") or settings.get("url")
        token = os.environ.get("NETBOX_TOKEN") or settings.get("token")

        if not url or not token:
            raise ValueError("Both url and token must be specified!")

        return pynetbox.api(url, token=token)

    def model_loader(self, model_name, model):
        for element in self.config.schema_mapping:
            if not element.name == model_name:
                continue

            app_name, resource_name = element.mapping.split(".")

            netbox_app = getattr(self.client, app_name)
            netbox_model = getattr(netbox_app, resource_name)

            objs = netbox_model.all()
            print(f"{self.type}: Loading {len(objs)} {resource_name}")
            for obj in objs:
                data = self.netbox_obj_to_diffsync(obj=obj, mapping=element, model=model)
                item = model(**data)
                self.add(item)

    def netbox_obj_to_diffsync(self, obj: NetboxRecord, mapping: SchemaMappingModel, model: NetboxModel) -> dict:  # pylint: disable=too-many-branches
        data: Dict[str, Any] = {"local_id": str(obj.id)}

        for field in mapping.fields:  # pylint: disable=too-many-nested-blocks
            field_is_list = model.is_list(name=field.name)

            if field.static:
                data[field.name] = field.static
            elif not field_is_list and field.mapping and not field.reference:
                value = get_value(obj, field.mapping)
                if value is not None:
                    data[field.name] = value
            elif field_is_list and field.mapping and not field.reference:
                raise NotImplementedError(
                    "it's not supported yet to have an attribute of type list with a simple mapping"
                )

            elif field.mapping and field.reference:
                all_nodes_for_reference = self.store.get_all(model=field.reference)
                nodes = [item for item in all_nodes_for_reference]  # noqa: C416
                if not nodes and all_nodes_for_reference:
                    raise IndexError(
                        f"Unable to get '{field.mapping}' with '{field.reference}' reference from store."
                        f" The available models are {self.store.get_all_model_names()}"
                    )
                if not field_is_list:
                    if node := get_value(obj, field.mapping):
                        matching_nodes = []
                        node_id = getattr(node, "id", None)
                        matching_nodes = [item for item in nodes if item.local_id == str(node_id)]
                        if len(matching_nodes) == 0:
                            raise IndexError(f"Unable to locate the node {model} {node_id}")
                        node = matching_nodes[0]
                        data[field.name] = node.get_unique_id()

                else:
                    data[field.name] = []
                    for node in get_value(obj, field.mapping):
                        if not node:
                            continue
                        node_id = getattr(node, "id", None)
                        if not node_id:
                            if isinstance(node, tuple):
                                node_id = node[1] if node[0] == "id" else None
                                if not node_id:
                                    continue
                        matching_nodes = [item for item in nodes if item.local_id == str(node_id)]
                        if len(matching_nodes) == 0:
                            raise IndexError(f"Unable to locate the node {field.reference} {node_id}")
                        data[field.name].append(matching_nodes[0].get_unique_id())

        return data


class NetboxModel(DiffSyncModelMixin, DiffSyncModel):
    pass
